/*
 * libusb rockband 2 drum interface
 * based on U.are.U 4000B fingerprint scanner.
 * Copyright (C) 2007 Daniel Drake <dsd@gentoo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <libusb-1.0/libusb.h>

#include <alsa/asoundlib.h>
#include <alsa/seq.h>
#include "pattern.h"

#define EP_INTR			(1 | LIBUSB_ENDPOINT_IN)
//#define EP_DATA			(2 | LIBUSB_ENDPOINT_IN)
//#define CTRL_IN			(LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_ENDPOINT_IN)
//#define CTRL_OUT		(LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_ENDPOINT_OUT)
//#define USB_RQ			0x04
#define INTR_LENGTH		27

#define DEFAULT_CHANNEL 9

#define max(a, b) (((a) > (b)) ? (a) : (b))
#define min(a, b) (((a) < (b)) ? (a) : (b))

void notedown(snd_seq_t *seq, int port, int chan, int pitch, int vol);
void noteup(snd_seq_t *seq, int port, int chan, int pitch, int vol);

static snd_seq_t *g_seq;
static int g_port;

static int verbose = 0;
static int g_i = 0;
static int state = 0;
static int drum_state[8];
static int do_exit = 0;
static unsigned char irqbuf[INTR_LENGTH];
static struct libusb_device_handle *devh = NULL;
//static struct libusb_transfer *img_transfer = NULL;
static struct libusb_transfer *irq_transfer = NULL;

static int find_rbdrum_device(int i)
{
	// TODO: Currently the i argument is ignored.
	devh = libusb_open_device_with_vid_pid(NULL, 0x12ba, 0x0210);
	return devh ? 0 : -EIO;
}

/*static int print_f0_data(void)
{
	unsigned char data[0x10];
	int r;
	unsigned int i;

	r = libusb_control_transfer(devh, CTRL_IN, USB_RQ, 0xf0, 0, data,
		sizeof(data), 0);
	if (r < 0) {
		fprintf(stderr, "F0 error %d\n", r);
		return r;
	}
	if ((unsigned int) r < sizeof(data)) {
		fprintf(stderr, "short read (%d)\n", r);
		return -1;
	}

	printf("F0 data:");
	for (i = 0; i < sizeof(data); i++)
		printf("%02x ", data[i]);
	printf("\n");
	return 0;
}

static int set_hwstat(unsigned char data)
{
	int r;

	printf("set hwstat to %02x\n", data);
	r = libusb_control_transfer(devh, CTRL_OUT, USB_RQ, 0x07, 0, &data, 1, 0);
	if (r < 0) {
		fprintf(stderr, "set hwstat error %d\n", r);
		return r;
	}
	if ((unsigned int) r < 1) {
		fprintf(stderr, "short write (%d)", r);
		return -1;
	}

	return 0;
}

static int get_hwstat(unsigned char *status)
{
	int r;

	r = libusb_control_transfer(devh, CTRL_IN, USB_RQ, 0x07, 0, status, 1, 0);
	if (r < 0) {
		fprintf(stderr, "read hwstat error %d\n", r);
		return r;
	}
	if ((unsigned int) r < 1) {
		fprintf(stderr, "short read (%d)\n", r);
		return -1;
	}

	printf("hwstat reads %02x\n", *status);
	return 0;
}

static int set_mode(unsigned char data)
{
	int r;
	printf("set mode %02x\n", data);

	r = libusb_control_transfer(devh, CTRL_OUT, USB_RQ, 0x4e, 0, &data, 1, 0);
	if (r < 0) {
		fprintf(stderr, "set mode error %d\n", r);
		return r;
	}
	if ((unsigned int) r < 1) {
		fprintf(stderr, "short write (%d)", r);
		return -1;
	}

	return 0;
}

static void cb_mode_changed(struct libusb_transfer *transfer)
{
	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		fprintf(stderr, "mode change transfer not completed!\n");
		do_exit = 2;
	}

	printf("async cb_mode_changed length=%d actual_length=%d\n",
		transfer->length, transfer->actual_length);
	if (next_state() < 0)
		do_exit = 2;
}

static int set_mode_async(unsigned char data)
{
	unsigned char *buf = malloc(LIBUSB_CONTROL_SETUP_SIZE + 1);
	struct libusb_transfer *transfer;

	if (!buf)
		return -ENOMEM;
	
	transfer = libusb_alloc_transfer(0);
	if (!transfer) {
		free(buf);
		return -ENOMEM;
	}

	printf("async set mode %02x\n", data);
	libusb_fill_control_setup(buf, CTRL_OUT, USB_RQ, 0x4e, 0, 1);
	buf[LIBUSB_CONTROL_SETUP_SIZE] = data;
	libusb_fill_control_transfer(transfer, devh, buf, cb_mode_changed, NULL,
		1000);

	transfer->flags = LIBUSB_TRANSFER_SHORT_NOT_OK
		| LIBUSB_TRANSFER_FREE_BUFFER | LIBUSB_TRANSFER_FREE_TRANSFER;
	return libusb_submit_transfer(transfer);
}*/

static int do_sync_intr(unsigned char *data)
{
	int r;
	int transferred;

	r = libusb_interrupt_transfer(devh, EP_INTR, data, INTR_LENGTH,
		&transferred, 1000);
	if (r < 0) {
		fprintf(stderr, "intr error %d\n", r);
		return r;
	}
	if (transferred < INTR_LENGTH) {
		fprintf(stderr, "short read (%d)\n", r);
		return -1;
	}

	printf("recv interrupt %04x\n", *((uint16_t *) data));
	return 0;
}

static int sync_intr(unsigned char type)
{	
	int r;
	unsigned char data[INTR_LENGTH];

	while (1) {
		r = do_sync_intr(data);
		if (r < 0)
			return r;
		if (data[0] == type)
			return 0;
	}
}

/*static int next_state(void)
{
	int r = 0;
	printf("old state: %d\n", state);
	switch (state) {
	case STATE_AWAIT_IRQ_FINGER_REMOVED:
		state = STATE_AWAIT_MODE_CHANGE_AWAIT_FINGER_ON;
		r = set_mode_async(MODE_AWAIT_FINGER_ON);
		break;
	case STATE_AWAIT_MODE_CHANGE_AWAIT_FINGER_ON:
		state = STATE_AWAIT_IRQ_FINGER_DETECTED;
		break;
	case STATE_AWAIT_IRQ_FINGER_DETECTED:
		state = STATE_AWAIT_MODE_CHANGE_CAPTURE;
		r = set_mode_async(MODE_CAPTURE);
		break;
	case STATE_AWAIT_MODE_CHANGE_CAPTURE:
		state = STATE_AWAIT_IMAGE;
		break;
	case STATE_AWAIT_IMAGE:
		state = STATE_AWAIT_MODE_CHANGE_AWAIT_FINGER_OFF;
		r = set_mode_async(MODE_AWAIT_FINGER_OFF);
		break;
	case STATE_AWAIT_MODE_CHANGE_AWAIT_FINGER_OFF:
		state = STATE_AWAIT_IRQ_FINGER_REMOVED;
		break;
	default:
		printf("unrecognised state %d\n", state);
	}
	if (r < 0) {
		fprintf(stderr, "error detected changing state\n");
		return r;
	}

	printf("new state: %d\n", state);
	return 0;
}*/

static void cb_irq(struct libusb_transfer *transfer)
{
	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		fprintf(stderr, "irq transfer status %d? %d\n", transfer->status, LIBUSB_TRANSFER_ERROR);
		do_exit = 2;
		libusb_free_transfer(transfer);
		irq_transfer = NULL;
		return;
	}

	// 60 = cowbell = blue
	// 57 = cymbal = green
	// 69 = better hi hat = yellow
	// 48 = snare = red
	// 35 = bass = bass
	// 64 = tom = bass
	// 59 = hi hat = yellow
	//int drum_conn[] = {60, 57, 48, 69, 35, };
	
	// 38 acoustic snare
	// 47 low-mid tom
	// 48 hi-mid tom
	// 50 hi tom
	// 44 pedal hi-hat
	// 35 acoustic bass drum
	// 56 cowbell
	// 49 cymbal
	int drum_conn[] = {56, 49, 38, 44, 35, 47, 48, 50};
	
	//                         hh sncyco
	// 0000087f 7f7f7f00 00000000 00000000 00000002 00020002 000200
	unsigned char *buf = irqbuf;
	unsigned char hihat   = buf[11];
	unsigned char snare   = buf[12];
	unsigned char cymbal  = buf[13];
	unsigned char cowbell = buf[14];
	unsigned char tom1    = 0;       // Yellow tom, see below
	unsigned char tom2    = 0;       // Blue   tom, see below
	unsigned char tom3    = 0;       // Green  tom, see below
	unsigned char bass    = buf[0] & 0x10;
	unsigned char bass_down = 0;

	if (verbose && (g_i++ % 200 == 199)) {
		printf("%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x\n", 
		buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], buf[8], buf[9], 
		buf[10], buf[11], buf[12], buf[13], buf[14], buf[15], buf[16], buf[17], buf[18], buf[19], 
		buf[20], buf[21], buf[22], buf[23], buf[24], buf[25], buf[26]);
	}
	if(hihat && buf[1] == 4)
	{
		tom1 = hihat;
		hihat = 0;
	}
	if(cowbell && buf[1] == 4)
	{
		tom2 = cowbell;
		cowbell = 0;
	}
	if(cymbal && buf[1] == 4)
	{
		tom3 = cymbal;
		cymbal = 0;
	}
	int velocity;
	if (snare || drum_state[0]) {
		// Events:
		// Down
		if (snare && !drum_state[0]) {
			velocity = 280-snare;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[2], velocity);
		}
		// Up
		if (!snare && drum_state[0]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[2], 0);
		}
		drum_state[0] = snare;
	}
	if (hihat || drum_state[1]) {
		// Events:
		// Down
		if (hihat && !drum_state[1]) {
			velocity = 280-hihat;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[3], velocity);
		}
		// Up
		if (!hihat && drum_state[1]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[3], 0);
		}
		drum_state[1] = hihat;
	}
	if (cowbell || drum_state[2]) {
		// Events:
		// Down
		if (cowbell && !drum_state[2]) {
			velocity = 280-cowbell;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[0], velocity);
		}
		// Up
		if (!cowbell && drum_state[2]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[0], 0);
		}
		drum_state[2] = cowbell;
	}
	if (cymbal || drum_state[3]) {
		// Events:
		// Down
		if (cymbal && !drum_state[3]) {
			velocity = 280-cymbal;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[1], velocity);
		}
		// Up
		if (!cymbal && drum_state[3]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[1], 0);
		}
		drum_state[3] = cymbal;
	}
	
	if (bass || drum_state[4]) {
		// Events:
		// Down
		if (bass && !drum_state[4]) {
			bass_down = 1;
			velocity = 127;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[4], velocity);
		}
		// Up
		if (!bass && drum_state[4]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[4], 0);
		}
		drum_state[4] = bass;
	}
	// Yellow Tom 1
	if (tom1 || drum_state[5]) {
		// Events:
		// Down
		if (tom1 && !drum_state[5]) {
			velocity = 280-tom1;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[5], velocity);
		}
		// Up
		if (!tom1 && drum_state[5]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[5], 0);
		}
		drum_state[5] = tom1;
	}
	// Blue Tom 2
	if (tom2 || drum_state[6]) {
		// Events:
		// Down
		if (tom2 && !drum_state[6]) {
			velocity = 280-tom2;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[6], velocity);
		}
		// Up
		if (!tom2 && drum_state[6]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[6], 0);
		}
		drum_state[6] = tom2;
	}
	// Yellow Tom 1
	if (tom3 || drum_state[7]) {
		// Events:
		// Down
		if (tom3 && !drum_state[7]) {
			velocity = 280-tom3;
			velocity = min(max(velocity, 0), 127);
			notedown(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[7], velocity);
		}
		// Up
		if (!tom3 && drum_state[7]) {
			noteup(g_seq, g_port, DEFAULT_CHANNEL, drum_conn[7], 0);
		}
		drum_state[7] = tom3;
	}
	if (verbose && (snare || hihat || cowbell || cymbal || bass_down || tom1 || tom2 || tom3)) {
		printf("%s %s %s %s %s %s %s\n", snare>0?"VV":"  ", hihat>0?"VV":"  ", tom1>0?"VV":"  ", cowbell>0?"VV":"  ", tom2>0?"VV":"  ", cymbal>0?"VV":"  ", tom3>0?"VV":"  ");
		printf("%02i %02i %02i %02i %02i %02i %02i\n", snare, hihat, tom1, cowbell, tom2, cymbal, tom3);
		printf("%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x\n", 
		buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], buf[8], buf[9], 
		buf[10], buf[11], buf[12], buf[13], buf[14], buf[15], buf[16], buf[17], buf[18], buf[19], 
		buf[20], buf[21], buf[22], buf[23], buf[24], buf[25], buf[26]);
	}

	if (libusb_submit_transfer(irq_transfer) < 0)
		do_exit = 2;
}

static int init_capture(void)
{
	int r;

	r = libusb_submit_transfer(irq_transfer);
	if (r < 0)
		return r;

	/*r = libusb_submit_transfer(img_transfer);
	if (r < 0) {
		libusb_cancel_transfer(irq_transfer);
		while (irq_transfer)
			if (libusb_handle_events(NULL) < 0)
				break;
		return r;
	}*/

	/* start state machine */
	state = 6; //STATE_AWAIT_IRQ_FINGER_REMOVED;
	return state; //next_state();
}

static int do_init(void)
{
	//unsigned char status;
	int r;

	/*r = get_hwstat(&status);
	if (r < 0)
		return r;

	if (!(status & 0x80)) {
		r = set_hwstat(status | 0x80);
		if (r < 0)
			return r;
		r = get_hwstat(&status);
		if (r < 0)
			return r;
	}

	status &= ~0x80;
	r = set_hwstat(status);
	if (r < 0)
		return r;

	r = get_hwstat(&status);
	if (r < 0)
		return r;

	r = sync_intr(0x56);
	if (r < 0)
		return r;*/

	return 0;
}

static int alloc_transfers(void)
{
	/*img_transfer = libusb_alloc_transfer(0);
	if (!img_transfer)
		return -ENOMEM;*/
	
	irq_transfer = libusb_alloc_transfer(0);
	if (!irq_transfer)
		return -ENOMEM;

	/*libusb_fill_bulk_transfer(img_transfer, devh, EP_DATA, imgbuf,
		sizeof(imgbuf), cb_img, NULL, 0);*/
	libusb_fill_interrupt_transfer(irq_transfer, devh, EP_INTR, irqbuf,
		sizeof(irqbuf), cb_irq, NULL, 0);
	

	return 0;
}

/*
These functions are from the official alsa docs:
http://www.alsa-project.org/alsa-doc/alsa-lib/seq.html
*/

// create a new client
snd_seq_t *open_client()
{
	snd_seq_t *handle;
	int err;
	err = snd_seq_open(&handle, "default", SND_SEQ_OPEN_OUTPUT, 0);
	if (err < 0)
		return NULL;
	snd_seq_set_client_name(handle, "PS3 Joystick Client");
	return handle;
}

// create a new port; return the port id
// port will be writable and accept the write-subscription.
int my_new_port(snd_seq_t *handle)
{
	// |SND_SEQ_PORT_CAP_WRITE||SND_SEQ_PORT_CAP_SUBS_WRITE
	return snd_seq_create_simple_port(handle, "PS3 Joystick port 2",
		SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
		SND_SEQ_PORT_TYPE_MIDI_GENERIC | SND_SEQ_PORT_TYPE_APPLICATION);
}

// create a queue and return its id
int my_queue(snd_seq_t *handle)
{
	return snd_seq_alloc_named_queue(handle, "PS3 Joystick queue");
}

// set the tempo and pulse per quarter note
void set_tempo(snd_seq_t *handle, int q)
{
	snd_seq_queue_tempo_t *tempo;
	snd_seq_queue_tempo_alloca(&tempo);
	perror("snd_seq_queue_tempo_alloca");
	snd_seq_queue_tempo_set_tempo(tempo, 1000000); // 60 BPM
	perror("snd_seq_queue_tempo_set_tempo");
	snd_seq_queue_tempo_set_ppq(tempo, 48); // 48 PPQ
	perror("snd_seq_queue_tempo_set_ppq");
	snd_seq_set_queue_tempo(handle, q, tempo);
	perror("snd_seq_set_queue_tempo");
}

// change the tempo on the fly
int change_tempo(snd_seq_t *handle, int my_client_id, int my_port_id, int q, unsigned int tempo)
{
	snd_seq_event_t ev;
	snd_seq_ev_clear(&ev);
	ev.dest.client = SND_SEQ_CLIENT_SYSTEM;
	ev.dest.port = SND_SEQ_PORT_SYSTEM_TIMER;
	ev.source.client = my_client_id;
	ev.source.port = my_port_id;
	ev.queue = SND_SEQ_QUEUE_DIRECT; // no scheduling
	ev.data.queue.queue = q;        // affected queue id
	ev.data.queue.param.value = tempo;    // new tempo in microsec.
	return snd_seq_event_output(handle, &ev);
}

static void program_change(snd_seq_t *seq, int port, int chan, int program)
{
	snd_seq_event_t ev;

	snd_seq_ev_clear(&ev);
	snd_seq_ev_set_source(&ev, port);
	snd_seq_ev_set_subs(&ev);
	snd_seq_ev_set_direct(&ev);
	//... // set event type, data, so on..
	//set_event_time(&ev, Mf_currtime);
	//snd_seq_ev_schedule_tick(&ev, q, 0, Mf_currtime);

	snd_seq_ev_set_pgmchange(&ev, chan, program);
	
	int rc = snd_seq_event_output(seq, &ev);
	if (rc < 0) {
		printf("written = %i (%s)\n", rc, snd_strerror(rc));
	}
	snd_seq_drain_output(seq);
}

// A lot easier:
void subscribe_output(snd_seq_t *seq, int client, int port)
{
	snd_seq_connect_to(seq, DEFAULT_CHANNEL, client, port);
}

/*// send an event to our subscriber.
// since we want realtime instead of tick clock, we replace the function.
void schedule_event(snd_seq_t *seq, int my_port_id, int q)
{
	snd_seq_event_t ev;

	snd_seq_ev_clear(&ev);
	snd_seq_ev_set_source(&ev, my_port_id);
	snd_seq_ev_set_subs(&ev);
	// this t makes no sense...
	snd_seq_real_time_t t;
	// Use realtime instead of tick clock.
	snd_seq_ev_schedule_real(&ev, q, 0, &t);
	//snd_seq_ev_schedule_tick(&ev, Q, 0, t);
	//... TODO:
	// set event type, data, so on..
	snd_seq_ev_clear(&ev);
	
	ev.queue = SND_SEQ_QUEUE_DIRECT; // no scheduling
	ev.data.queue.queue = q;        // affected queue id
	//ev.data.queue.param.value = ...

	snd_seq_event_output(seq, &ev);
	//... TODO: ?
	snd_seq_drain_output(seq);  // if necessary
}*/

// From test/playmidi1.c from alsa-lib-1.0.3.

// Direct delivery seems like what I'm doing..
void notedown(snd_seq_t *seq, int port, int chan, int pitch, int vol)
{
	snd_seq_event_t ev;

	snd_seq_ev_clear(&ev);
	snd_seq_ev_set_source(&ev, port);
	snd_seq_ev_set_subs(&ev);
	snd_seq_ev_set_direct(&ev);
	//... // set event type, data, so on..
	//set_event_time(&ev, Mf_currtime);
	//snd_seq_ev_schedule_tick(&ev, q, 0, Mf_currtime);

	snd_seq_ev_set_noteon(&ev, chan, pitch, vol);
	
	int rc = snd_seq_event_output(seq, &ev);
	if (rc < 0) {
		printf("written = %i (%s)\n", rc, snd_strerror(rc));
	}
	snd_seq_drain_output(seq);
}

// When the note up, note off.
void noteup(snd_seq_t *seq, int port, int chan, int pitch, int vol)
{
	snd_seq_event_t ev;

	snd_seq_ev_clear(&ev);
	snd_seq_ev_set_source(&ev, port);
	snd_seq_ev_set_subs(&ev);
	snd_seq_ev_set_direct(&ev);
	//... // set event type, data, so on..

	snd_seq_ev_set_noteoff(&ev, chan, pitch, vol);
	
	snd_seq_event_output(seq, &ev);
	snd_seq_drain_output(seq);
}

int setup_alsa(snd_seq_t **seq, int *port)
{
	if (verbose) printf("Setting up alsa\n");
	
	*seq = open_client();
	if (*seq == NULL) {
		if (verbose >= 0) printf("Error: open_client failed: %s\n", snd_strerror(seq));
		return 0;
	}
	
	int my_client_id = snd_seq_client_id(*seq);
	*port = my_new_port(*seq);
	if (verbose) printf("client:port = %i:%i\n", my_client_id, *port);

	program_change(*seq, *port, DEFAULT_CHANNEL, 0);
	// Subscribing to something else seems unnecssary until we figure out more.
	//subscribe_output(seq, 128, 0);
	
	//int tempo = 120;
	//int q = my_queue(seq);
	int ret = 1;
	/*
	// tempo isn't being used yet, errors *shrug*.
	set_tempo(seq, q);
	int ret = change_tempo(seq, my_client_id, my_port_id, q, tempo);
	if (ret < 0) {
		if (verbose >= 0) printf("Error: change_tempo failed: %s\n", snd_strerror(ret));
		return ret;
	}
	*/
	// Scheduling is scheduling, direct delivery is what I'm doing.
	//schedule_event(seq, my_port_id, q);
	notedown(*seq, *port, DEFAULT_CHANNEL, 57, 55);
	
	if (verbose) printf("Returning %i\n", ret);
	return ret;
}

void testAlsa(snd_seq_t *seq, int port)
{
	notedown(seq, port, DEFAULT_CHANNEL, 57, 127);
	usleep(1000000);
	noteup(seq, port, DEFAULT_CHANNEL, 57, 0);
	usleep(1000000);

}

static void sighandler(int signum)
{
	do_exit = 1;
}

int main(int argc, char **argv)
{
	struct sigaction sigact;
	int r = 1;

	r = libusb_init(NULL);
	if (r < 0) {
		fprintf(stderr, "failed to initialise libusb\n");
		exit(1);
	}
	int i = 0;

	// TODO: Currently this argument is ignored.
	if (argc > 1) {
		if (strcmp(argv[1], "-v") == 0) {
			verbose = 1;
		}
		i = atoi(argv[1]);
	}

	r = find_rbdrum_device(i);
	if (r < 0) {
		fprintf(stderr, "Could not find/open device\n");
		libusb_close(devh);
		libusb_exit(NULL);
		return -r;
	}

	if (libusb_kernel_driver_active(devh, 0)) {
		/*char driver_name[100];
		driver_name[0] = 0;
		r = usb_get_driver_np(devh, 0, driver_name, 100);
		if (r < 0) {
			printf("did not get driver_name.\n");
			driver_name[0] = 0;
		}*/
		r = libusb_detach_kernel_driver(devh, 0);
		if (r < 0) {
			printf("did not detach.\n");
		}
	}
	
	r = libusb_claim_interface(devh, 0);
	if (r < 0) {
		fprintf(stderr, "usb_claim_interface error %d %d\n", r, LIBUSB_ERROR_BUSY);
		libusb_close(devh);
		libusb_exit(NULL);
		return -r;
	}
	printf("claimed interface\n");

	int ret = setup_alsa(&g_seq, &g_port);
	// 0 is fail.
	if (ret == 0) {
		printf("Error: Alsa setup failed.\n");
		return 1;
	}

	/*r = print_f0_data();
	if (r < 0) {
		// Release
		libusb_release_interface(devh, 0);
		libusb_close(devh);
		libusb_exit(NULL);
		return -r;
	}*/

	r = do_init();
	if (r < 0) {
		// Deinit & Release
		libusb_free_transfer(irq_transfer);
		libusb_release_interface(devh, 0);
		libusb_close(devh);
		libusb_exit(NULL);
		snd_seq_close(g_seq);
		printf("do_init failed.\n");
		return -r;
	}

	/* async from here onwards */

	r = alloc_transfers();
	if (r < 0) {
		// Deinit & Release
		libusb_free_transfer(irq_transfer);
		libusb_release_interface(devh, 0);
		libusb_close(devh);
		libusb_exit(NULL);
		snd_seq_close(g_seq);
		printf("alloc_transfers failed.\n");
		return -r;
	}

	r = init_capture();
	if (r < 0) {
		// Deinit & Release
		libusb_free_transfer(irq_transfer);
		libusb_release_interface(devh, 0);
		libusb_close(devh);
		libusb_exit(NULL);
		snd_seq_close(g_seq);
		printf("init_capture failed.\n");
		return -r;
	}

	sigact.sa_handler = sighandler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = SA_RESTART;
	r = sigaction(SIGINT, &sigact, NULL);
	r = sigaction(SIGTERM, &sigact, NULL);
	r = sigaction(SIGQUIT, &sigact, NULL);

	// Drum state is all up.
	memset(drum_state, 0, 5);

	while (!do_exit) {
		r = libusb_handle_events(NULL);
		if (r < 0) {
			/*printf("why?\n");
			// Deinit & Release
			libusb_free_transfer(irq_transfer);
			libusb_release_interface(devh, 0);
			libusb_close(devh);
			libusb_exit(NULL);
			return -r;*/
			break;
		}
	}

	printf("shutting down...\n");
	
	if (irq_transfer) {
		r = libusb_cancel_transfer(irq_transfer);
		if (r < 0) {
			// Deinit & Release
			libusb_free_transfer(irq_transfer);
			libusb_release_interface(devh, 0);
			libusb_close(devh);
			libusb_exit(NULL);
			snd_seq_close(g_seq);
			printf("libusb_cancel_transfer failed.\n");
			return -r;
		}
	}

	/*if (img_transfer) {
		r = libusb_cancel_transfer(img_transfer);
		if (r < 0) {
			// TODO: Deinit & Release
			libusb_release_interface(devh, 0);
			libusb_close(devh);
			libusb_exit(NULL);
			return -r;
		}
	}*/

	// || img_transfer
	while (irq_transfer)
		if (libusb_handle_events(NULL) < 0)
			break;
	
	if (do_exit == 1)
		r = 0;
	else
		r = 1;

//out_deinit:
	//libusb_free_transfer(img_transfer);
	libusb_free_transfer(irq_transfer);
	/*set_mode(0);
	set_hwstat(0x80);*/
//out_release:
	libusb_release_interface(devh, 0);
//out:
	libusb_close(devh);
	libusb_exit(NULL);
	snd_seq_close(g_seq);
	return 23;
	return r >= 0 ? r : -r;
}
